import functools
import json
from pathlib import (
    Path,
)
import tempfile
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
)

from eth_typing import (
    URI,
    HexStr,
    Manifest,
)
from eth_utils import (
    add_0x_prefix,
    is_hex,
    is_string,
    to_bytes,
    to_checksum_address,
    to_dict,
    to_list,
)
from eth_utils.toolz import (
    assoc,
    assoc_in,
    concat,
    curry,
    pipe,
)

from ethpm import (
    Package,
)
from ethpm._utils.chains import (
    is_BIP122_block_uri,
)
from ethpm.backends.ipfs import (
    BaseIPFSBackend,
)
from ethpm.exceptions import (
    EthPMValidationError,
    ManifestBuildingError,
)
from ethpm.package import (
    format_manifest,
)
from ethpm.uri import (
    is_supported_content_addressed_uri,
)
from ethpm.validation.manifest import (
    validate_manifest_against_schema,
)
from ethpm.validation.package import (
    validate_package_name,
)
from newchain_web3._utils.validation import (
    validate_address,
)

if TYPE_CHECKING:
    from newchain_web3 import Web3  # noqa: F401


def build(obj: Dict[str, Any], *fns: Callable[..., Any]) -> Dict[str, Any]:
    """
    Wrapper function to pipe manifest through build functions.
    Does not validate the manifest by default.
    """
    return pipe(obj, *fns)


#
# Required Fields
#


@curry
def package_name(name: str, manifest: Manifest) -> Manifest:
    """
    Return a copy of manifest with `name` set to "name".
    """
    return assoc(manifest, "name", name)


@curry
def manifest_version(manifest_version: str, manifest: Manifest) -> Manifest:
    """
    Return a copy of manifest with `manifest_version` set to "manifest".
    """
    return assoc(manifest, "manifest", manifest_version)


@curry
def version(version: str, manifest: Manifest) -> Manifest:
    """
    Return a copy of manifest with `version` set to "version".
    """
    return assoc(manifest, "version", version)


#
# Meta
#


def authors(*author_list: str) -> Manifest:
    """
    Return a copy of manifest with a list of author posargs set
    to "meta": {"authors": author_list}
    """
    return _authors(author_list)


@curry
@functools.wraps(authors)
def _authors(authors: Set[str], manifest: Manifest) -> Manifest:
    return assoc_in(manifest, ("meta", "authors"), list(authors))


@curry
def license(license: str, manifest: Manifest) -> Manifest:
    """
    Return a copy of manifest with `license` set to
    "meta": {"license": `license`}
    """
    return assoc_in(manifest, ("meta", "license"), license)


@curry
def description(description: str, manifest: Manifest) -> Manifest:
    """
    Return a copy of manifest with `description` set to
    "meta": {"descriptions": `description`}
    """
    return assoc_in(manifest, ("meta", "description"), description)


def keywords(*keyword_list: str) -> Manifest:
    """
    Return a copy of manifest with a list of keyword posargs set to
    "meta": {"keywords": keyword_list}
    """
    return _keywords(keyword_list)


@curry
@functools.wraps(keywords)
def _keywords(keywords: Set[str], manifest: Manifest) -> Manifest:
    return assoc_in(manifest, ("meta", "keywords"), list(keywords))


def links(**link_dict: str) -> Manifest:
    """
    Return a copy of manifest with a dict of link kwargs set to
    "meta": {"links": link_dict}
    """
    return _links(link_dict)


@curry
def _links(link_dict: Dict[str, str], manifest: Manifest) -> Manifest:
    return assoc_in(manifest, ("meta", "links"), link_dict)


#
# Sources
#


def get_names_and_paths(compiler_output: Dict[str, Any]) -> Dict[str, str]:
    """
    Return a mapping of contract name to relative path as defined in compiler output.
    """
    return {
        contract_name: make_path_relative(path)
        for path in compiler_output
        for contract_name in compiler_output[path].keys()
    }


def make_path_relative(path: str) -> str:
    """
    Returns the given path prefixed with "./" if the path
    is not already relative in the compiler output.
    """
    if "../" in path:
        raise ManifestBuildingError(
            f"Path: {path} appears to be outside of the virtual source tree. "
            "Please make sure all sources are within the virtual source tree "
            "root directory."
        )

    if path[:2] != "./":
        return f"./{path}"
    return path


def source_inliner(
    compiler_output: Dict[str, Any], package_root_dir: Optional[Path] = None
) -> Manifest:
    return _inline_sources(compiler_output, package_root_dir)


@curry
def _inline_sources(
    compiler_output: Dict[str, Any], package_root_dir: Optional[Path], name: str
) -> Manifest:
    return _inline_source(name, compiler_output, package_root_dir)


def inline_source(
    name: str, compiler_output: Dict[str, Any], package_root_dir: Optional[Path] = None
) -> Manifest:
    """
    Return a copy of manifest with added field to
    "sources": {relative_source_path: contract_source_data}.

    If `package_root_dir` is not provided, cwd is expected to resolve the relative
    path to the source as defined in the compiler output.
    """
    return _inline_source(name, compiler_output, package_root_dir)


@curry
def _inline_source(
    name: str,
    compiler_output: Dict[str, Any],
    package_root_dir: Optional[Path],
    manifest: Manifest,
) -> Manifest:
    names_and_paths = get_names_and_paths(compiler_output)
    cwd = Path.cwd()
    try:
        source_path = names_and_paths[name]
    except KeyError:
        raise ManifestBuildingError(
            f"Unable to inline source: {name}. "
            f"Available sources include: {list(sorted(names_and_paths.keys()))}."
        )

    if package_root_dir:
        if (package_root_dir / source_path).is_file():
            source_data = (package_root_dir / source_path).read_text()
        else:
            raise ManifestBuildingError(
                f"Contract source: {source_path} cannot be found in "
                f"provided package_root_dir: {package_root_dir}."
            )
    elif (cwd / source_path).is_file():
        source_data = (cwd / source_path).read_text()
    else:
        raise ManifestBuildingError(
            "Contract source cannot be resolved, please make sure that the working "
            "directory is set to the correct directory or provide `package_root_dir`."
        )

    # rstrip used here since Path.read_text() adds a newline to returned contents
    source_data_object = {
        "content": source_data.rstrip("\n"),
        "installPath": source_path,
        "type": "solidity",
    }
    return assoc_in(manifest, ["sources", source_path], source_data_object)


def source_pinner(
    compiler_output: Dict[str, Any],
    ipfs_backend: BaseIPFSBackend,
    package_root_dir: Optional[Path] = None,
) -> Manifest:
    return _pin_sources(compiler_output, ipfs_backend, package_root_dir)


@curry
def _pin_sources(
    compiler_output: Dict[str, Any],
    ipfs_backend: BaseIPFSBackend,
    package_root_dir: Optional[Path],
    name: str,
) -> Manifest:
    return _pin_source(name, compiler_output, ipfs_backend, package_root_dir)


def pin_source(
    name: str,
    compiler_output: Dict[str, Any],
    ipfs_backend: BaseIPFSBackend,
    package_root_dir: Optional[Path] = None,
) -> Manifest:
    """
    Pins source to IPFS and returns a copy of manifest with added field to
    "sources": {relative_source_path: IFPS URI}.

    If `package_root_dir` is not provided, cwd is expected to resolve the relative path
    to the source as defined in the compiler output.
    """
    return _pin_source(name, compiler_output, ipfs_backend, package_root_dir)


@curry
def _pin_source(
    name: str,
    compiler_output: Dict[str, Any],
    ipfs_backend: BaseIPFSBackend,
    package_root_dir: Optional[Path],
    manifest: Manifest,
) -> Manifest:
    names_and_paths = get_names_and_paths(compiler_output)
    try:
        source_path = names_and_paths[name]
    except KeyError:
        raise ManifestBuildingError(
            f"Unable to pin source: {name}. "
            f"Available sources include: {list(sorted(names_and_paths.keys()))}."
        )
    if package_root_dir:
        if not (package_root_dir / source_path).is_file():
            raise ManifestBuildingError(
                f"Unable to find and pin contract source: {source_path} "
                f"under specified package_root_dir: {package_root_dir}."
            )
        (ipfs_data,) = ipfs_backend.pin_assets(package_root_dir / source_path)
    else:
        cwd = Path.cwd()
        if not (cwd / source_path).is_file():
            raise ManifestBuildingError(
                f"Unable to find and pin contract source: {source_path} "
                f"current working directory: {cwd}."
            )
        (ipfs_data,) = ipfs_backend.pin_assets(cwd / source_path)

    source_data_object = {
        "urls": [f"ipfs://{ipfs_data['Hash']}"],
        "type": "solidity",
        "installPath": source_path,
    }
    return assoc_in(manifest, ["sources", source_path], source_data_object)


#
# Contract Types
#


def contract_type(
    name: str,
    compiler_output: Dict[str, Any],
    alias: Optional[str] = None,
    abi: Optional[bool] = False,
    compiler: Optional[bool] = False,
    contract_type: Optional[bool] = False,
    deployment_bytecode: Optional[bool] = False,
    devdoc: Optional[bool] = False,
    userdoc: Optional[bool] = False,
    source_id: Optional[bool] = False,
    runtime_bytecode: Optional[bool] = False,
) -> Manifest:
    """
    Returns a copy of manifest with added contract_data field as specified by kwargs.
    If no kwargs are present, all available contract_data found in the compiler output
    will be included.

    To include specific contract_data fields, add kwarg set to True (i.e. `abi=True`)
    To alias a contract_type, include a kwarg `alias` (i.e. `alias="OwnedAlias"`)
    If only an alias kwarg is provided, all available contract data will be included.
    Kwargs must match fields as defined in the EthPM Spec (except "alias") if user
    wants to include them in custom contract_type.
    """
    contract_type_fields = {
        "contractType": contract_type,
        "deploymentBytecode": deployment_bytecode,
        "runtimeBytecode": runtime_bytecode,
        "abi": abi,
        "compiler": compiler,
        "userdoc": userdoc,
        "devdoc": devdoc,
        "sourceId": source_id,
    }
    selected_fields = [k for k, v in contract_type_fields.items() if v]
    return _contract_type(name, compiler_output, alias, selected_fields)


@curry
def _contract_type(
    name: str,
    compiler_output: Dict[str, Any],
    alias: Optional[str],
    selected_fields: Optional[List[str]],
    manifest: Manifest,
) -> Manifest:
    contracts_by_name = normalize_compiler_output(compiler_output)
    try:
        all_type_data = contracts_by_name[name]
    except KeyError:
        raise ManifestBuildingError(
            f"Contract name: {name} not found in the provided compiler output."
        )
    if selected_fields:
        contract_type_data = filter_all_data_by_selected_fields(
            all_type_data, selected_fields
        )
    else:
        contract_type_data = all_type_data

    if "compiler" in contract_type_data:
        compiler_info = contract_type_data.pop("compiler")
        contract_type_ref = alias if alias else name
        manifest_with_compilers = add_compilers_to_manifest(
            compiler_info, contract_type_ref, manifest
        )
    else:
        manifest_with_compilers = manifest

    if alias:
        return assoc_in(
            manifest_with_compilers,
            ["contractTypes", alias],
            assoc(contract_type_data, "contractType", name),
        )
    return assoc_in(
        manifest_with_compilers, ["contractTypes", name], contract_type_data
    )


def add_compilers_to_manifest(
    compiler_info: Dict[str, Any], contract_type: str, manifest: Manifest
) -> Manifest:
    """
    Adds a compiler information object to a manifest's top-level `compilers`.
    """
    if "compilers" not in manifest:
        compiler_info["contractTypes"] = [contract_type]
        return assoc_in(manifest, ["compilers"], [compiler_info])

    updated_compiler_info = update_compilers_object(
        compiler_info, contract_type, manifest["compilers"]
    )
    return assoc_in(manifest, ["compilers"], updated_compiler_info)


@to_list
def update_compilers_object(
    new_compiler: Dict[str, Any],
    contract_type: str,
    previous_compilers: List[Dict[str, Any]],
) -> Iterable[Dict[str, Any]]:
    """
    Updates a manifest's top-level `compilers` with a new compiler information object.
    - If compiler version already exists, we just update the compiler's `contractTypes`
    """
    recorded_new_contract_type = False
    for compiler in previous_compilers:
        contract_types = compiler.pop("contractTypes")
        if contract_type in contract_types:
            raise ManifestBuildingError(
                f"Contract type: {contract_type} already referenced in `compilers`."
            )
        if compiler == new_compiler:
            contract_types.append(contract_type)
            recorded_new_contract_type = True
        compiler["contractTypes"] = contract_types
        yield compiler

    if not recorded_new_contract_type:
        new_compiler["contractTypes"] = [contract_type]
        yield new_compiler


@to_dict
def filter_all_data_by_selected_fields(
    all_type_data: Dict[str, Any], selected_fields: List[str]
) -> Iterable[Tuple[str, Any]]:
    """
    Raises exception if selected field data is not available in the contract type data
    automatically gathered by normalize_compiler_output. Otherwise, returns the data.
    """
    for field in selected_fields:
        if field in all_type_data:
            yield field, all_type_data[field]
        else:
            raise ManifestBuildingError(
                f"Selected field: {field} not available in data collected from "
                f"solc output: {list(sorted(all_type_data.keys()))}. Please make"
                "sure the relevant data is present in your solc output."
            )


def normalize_compiler_output(compiler_output: Dict[str, Any]) -> Dict[str, Any]:
    """
    Return compiler output with normalized fields for each contract type,
    as specified in `normalize_contract_type`.
    """
    paths_and_names = [
        (path, contract_name)
        for path in compiler_output
        for contract_name in compiler_output[path].keys()
    ]
    paths, names = zip(*paths_and_names)
    if len(names) != len(set(names)):
        duplicates = {name for name in names if names.count(name) > 1}
        raise ManifestBuildingError(
            f"Duplicate contract types: {duplicates} were found in the compiler output."
        )
    return {
        name: normalize_contract_type(compiler_output[path][name], path)
        for path, name in paths_and_names
    }


@to_dict
def normalize_contract_type(
    contract_type_data: Dict[str, Any],
    source_id: str,
) -> Iterable[Tuple[str, Any]]:
    """
    Serialize contract_data found in compiler output to the defined fields.
    """
    yield "abi", contract_type_data["abi"]
    yield "sourceId", source_id
    if "evm" in contract_type_data:
        if "bytecode" in contract_type_data["evm"]:
            yield "deploymentBytecode", normalize_bytecode_object(
                contract_type_data["evm"]["bytecode"]
            )
        if "deployedBytecode" in contract_type_data["evm"]:
            yield "runtimeBytecode", normalize_bytecode_object(
                contract_type_data["evm"]["deployedBytecode"]
            )
    if "devdoc" in contract_type_data:
        yield "devdoc", contract_type_data["devdoc"]
    if "userdoc" in contract_type_data:
        yield "userdoc", contract_type_data["userdoc"]
    # make sure metadata isn't an empty string in solc output
    if "metadata" in contract_type_data and contract_type_data["metadata"]:
        yield "compiler", normalize_compiler_object(
            json.loads(contract_type_data["metadata"])
        )


@to_dict
def normalize_compiler_object(obj: Dict[str, Any]) -> Iterable[Tuple[str, Any]]:
    yield "name", "solc"
    yield "version", obj["compiler"]["version"]
    yield "settings", {"optimize": obj["settings"]["optimizer"]["enabled"]}


@to_dict
def normalize_bytecode_object(obj: Dict[str, Any]) -> Iterable[Tuple[str, Any]]:
    try:
        link_references = obj["linkReferences"]
    except KeyError:
        link_references = None
    try:
        bytecode = obj["object"]
    except KeyError:
        raise ManifestBuildingError(
            "'object' key not found in bytecode data from compiler output. "
            "Please make sure your solidity compiler output is valid."
        )
    if link_references:
        yield "linkReferences", process_link_references(link_references, bytecode)
        yield "bytecode", process_bytecode(link_references, bytecode)
    else:
        yield "bytecode", add_0x_prefix(bytecode)


def process_bytecode(link_refs: Dict[str, Any], bytecode: bytes) -> HexStr:
    """
    Replace link_refs in bytecode with 0's.
    """
    all_offsets = [y for x in link_refs.values() for y in x.values()]
    # Link ref validation.
    validate_link_ref_fns = (
        validate_link_ref(ref["start"] * 2, ref["length"] * 2)
        for ref in concat(all_offsets)
    )
    pipe(bytecode, *validate_link_ref_fns)
    # Convert link_refs in bytecode to 0's
    link_fns = (
        replace_link_ref_in_bytecode(ref["start"] * 2, ref["length"] * 2)
        for ref in concat(all_offsets)
    )
    processed_bytecode = pipe(bytecode, *link_fns)
    return add_0x_prefix(processed_bytecode)


@curry
def replace_link_ref_in_bytecode(offset: int, length: int, bytecode: str) -> str:
    new_bytes = (
        bytecode[:offset] + "0" * length + bytecode[offset + length :]  # noqa: E203
    )
    return new_bytes


# todo pull all bytecode linking/validating across py-ethpm into shared utils
@to_list
def process_link_references(
    link_refs: Dict[str, Any], bytecode: str
) -> Iterable[Dict[str, Any]]:
    for link_ref in link_refs.values():
        yield normalize_link_ref(link_ref, bytecode)


def normalize_link_ref(link_ref: Dict[str, Any], bytecode: str) -> Dict[str, Any]:
    name = list(link_ref.keys())[0]
    return {
        "name": name,
        "length": 20,
        "offsets": normalize_offsets(link_ref, bytecode),
    }


@to_list
def normalize_offsets(data: Dict[str, Any], bytecode: str) -> Iterable[List[int]]:
    for link_ref in data.values():
        for ref in link_ref:
            yield ref["start"]


@curry
def validate_link_ref(offset: int, length: int, bytecode: str) -> str:
    slot_length = offset + length
    slot = bytecode[offset:slot_length]
    if slot[:2] != "__" and slot[-2:] != "__":
        raise EthPMValidationError(
            f"Slot: {slot}, at offset: {offset} of length: {length} is not a valid "
            "link_ref that can be replaced."
        )
    return bytecode


#
# Deployments
#


def deployment_type(
    *,
    contract_instance: str,
    contract_type: str,
    deployment_bytecode: Dict[str, Any] = None,
    runtime_bytecode: Dict[str, Any] = None,
    compiler: Dict[str, Any] = None,
) -> Manifest:
    """
    Returns a callable that allows the user to add deployments of the same type
    across multiple chains.
    """
    return _deployment_type(
        contract_instance,
        contract_type,
        deployment_bytecode,
        runtime_bytecode,
        compiler,
    )


def deployment(
    *,
    block_uri: URI,
    contract_instance: str,
    contract_type: str,
    address: HexStr,
    transaction: HexStr = None,
    block: HexStr = None,
    deployment_bytecode: Dict[str, Any] = None,
    runtime_bytecode: Dict[str, Any] = None,
    compiler: Dict[str, Any] = None,
) -> Manifest:
    """
    Returns a manifest, with the newly included deployment. Requires a valid
    blockchain URI, however no validation is provided that this URI is unique
    amongst the other deployment URIs, so the user must take care that each
    blockchain URI represents a unique blockchain.
    """
    return _deployment(
        contract_instance,
        contract_type,
        deployment_bytecode,
        runtime_bytecode,
        compiler,
        block_uri,
        address,
        transaction,
        block,
    )


@curry
def _deployment_type(
    contract_instance: str,
    contract_type: str,
    deployment_bytecode: Dict[str, Any],
    runtime_bytecode: Dict[str, Any],
    compiler: Dict[str, Any],
    block_uri: URI,
    address: HexStr,
    tx: HexStr = None,
    block: HexStr = None,
    manifest: Manifest = None,
) -> Manifest:
    return _deployment(
        contract_instance,
        contract_type,
        deployment_bytecode,
        runtime_bytecode,
        compiler,
        block_uri,
        address,
        tx,
        block,
    )


@curry
def _deployment(
    contract_instance: str,
    contract_type: str,
    deployment_bytecode: Dict[str, Any],
    runtime_bytecode: Dict[str, Any],
    compiler: Dict[str, Any],
    block_uri: URI,
    address: HexStr,
    tx: HexStr,
    block: HexStr,
    manifest: Manifest,
) -> Manifest:
    validate_address(address)
    if not is_BIP122_block_uri(block_uri):
        raise ManifestBuildingError(f"{block_uri} is not a valid BIP122 URI.")

    if tx:
        if not is_string(tx) and not is_hex(tx):
            raise ManifestBuildingError(
                f"Transaction hash: {tx} is not a valid hexstring"
            )

    if block:
        if not is_string(block) and not is_hex(block):
            raise ManifestBuildingError(f"Block hash: {block} is not a valid hexstring")
    # todo: validate db, rb and compiler are properly formatted dicts
    deployment_data = _build_deployments_object(
        contract_type,
        deployment_bytecode,
        runtime_bytecode,
        compiler,
        address,
        tx,
        block,
        manifest,
    )
    return assoc_in(
        manifest, ["deployments", block_uri, contract_instance], deployment_data
    )


@to_dict
def _build_deployments_object(
    contract_type: str,
    deployment_bytecode: Dict[str, Any],
    runtime_bytecode: Dict[str, Any],
    compiler: Dict[str, Any],
    address: HexStr,
    tx: HexStr,
    block: HexStr,
    manifest: Dict[str, Any],
) -> Iterable[Tuple[str, Any]]:
    """
    Returns a dict with properly formatted deployment data.
    """
    yield "contractType", contract_type
    yield "address", to_checksum_address(address)
    if deployment_bytecode:
        yield "deploymentBytecode", deployment_bytecode
    if compiler:
        yield "compiler", compiler
    if tx:
        yield "transaction", tx
    if block:
        yield "block", block
    if runtime_bytecode:
        yield "runtimeBytecode", runtime_bytecode


#
# Build Dependencies
#


def build_dependency(package_name: str, uri: URI) -> Manifest:
    """
    Returns the manifest with injected build dependency.
    """
    return _build_dependency(package_name, uri)


@curry
def _build_dependency(package_name: str, uri: URI, manifest: Manifest) -> Manifest:
    validate_package_name(package_name)
    if not is_supported_content_addressed_uri(uri):
        raise EthPMValidationError(
            f"{uri} is not a supported content-addressed URI. "
            "Currently only IPFS and Github blob uris are supported."
        )
    return assoc_in(manifest, ("buildDependencies", package_name), uri)


#
# Helpers
#


@curry
def init_manifest(
    package_name: str, version: str, manifest_version: Optional[str] = "ethpm/3"
) -> Dict[str, Any]:
    """
    Returns an initial dict with the minimal requried fields for a valid manifest.
    Should only be used as the first fn to be piped into a `build()` pipeline.
    """
    return {
        "name": package_name,
        "version": version,
        "manifest": manifest_version,
    }


#
# Formatting
#


@curry
def validate(manifest: Manifest) -> Manifest:
    """
    Return a validated manifest against the V2-specification schema.
    """
    validate_manifest_against_schema(manifest)
    return manifest


@curry
def as_package(w3: "Web3", manifest: Manifest) -> Package:
    """
    Return a Package object instantiated with the provided manifest and web3 instance.
    """
    return Package(manifest, w3)


def write_to_disk(
    manifest_root_dir: Optional[Path] = None,
    manifest_name: Optional[str] = None,
    prettify: Optional[bool] = False,
) -> Manifest:
    """
    Write the active manifest to disk
    Defaults
    - Writes manifest to cwd unless Path is provided as manifest_root_dir.
    - Writes manifest with a filename of Manifest[version].json unless a desired
    manifest name (which must end in json) is provided as manifest_name.
    - Writes the minified manifest version to disk unless prettify is set to True.
    """
    return _write_to_disk(manifest_root_dir, manifest_name, prettify)


@curry
def _write_to_disk(
    manifest_root_dir: Optional[Path],
    manifest_name: Optional[str],
    prettify: Optional[bool],
    manifest: Manifest,
) -> Manifest:
    if manifest_root_dir:
        if manifest_root_dir.is_dir():
            cwd = manifest_root_dir
        else:
            raise ManifestBuildingError(
                f"Manifest root directory: {manifest_root_dir} cannot be found, please "
                "provide a valid directory for writing the manifest to disk. "
                "(Path obj // leave manifest_root_dir blank to default to cwd)"
            )
    else:
        cwd = Path.cwd()

    if manifest_name:
        if not manifest_name.lower().endswith(".json"):
            raise ManifestBuildingError(
                f"Invalid manifest name: {manifest_name}. "
                "All manifest names must end in .json"
            )
        disk_manifest_name = manifest_name
    else:
        disk_manifest_name = manifest["version"] + ".json"

    contents = format_manifest(manifest, prettify=prettify)

    if (cwd / disk_manifest_name).is_file():
        raise ManifestBuildingError(
            f"Manifest: {disk_manifest_name} already exists in cwd: {cwd}"
        )
    (cwd / disk_manifest_name).write_text(contents)
    return manifest


@curry
def pin_to_ipfs(
    manifest: Manifest, *, backend: BaseIPFSBackend, prettify: Optional[bool] = False
) -> List[Dict[str, str]]:
    """
    Returns the IPFS pin data after pinning the manifest to the provided IPFS Backend.

    `pin_to_ipfs()` Should *always* be the last argument in a builder, as it will
    return the pin data and not the manifest.
    """
    contents = format_manifest(manifest, prettify=prettify)

    with tempfile.NamedTemporaryFile() as temp:
        temp.write(to_bytes(text=contents))
        temp.seek(0)
        return backend.pin_assets(Path(temp.name))



import pytest
from typing import (
    TYPE_CHECKING,
)

from eth_utils import (
    is_boolean,
    is_integer,
    is_string,
)

if TYPE_CHECKING:
    from newchain_web3 import Web3  # noqa: F401


class NetModuleTest:
    def test_net_version(self, w3: "Web3") -> None:
        version = w3.net.version

        assert is_string(version)
        assert version.isdigit()

    def test_net_listening(self, w3: "Web3") -> None:
        listening = w3.net.listening

        assert is_boolean(listening)

    def test_net_peer_count(self, w3: "Web3") -> None:
        peer_count = w3.net.peer_count

        assert is_integer(peer_count)


class AsyncNetModuleTest:
    @pytest.mark.asyncio
    async def test_net_version(self, async_w3: "Web3") -> None:
        version = await async_w3.async_net.version

        assert is_string(version)
        assert version.isdigit()

    @pytest.mark.asyncio
    async def test_net_listening(self, async_w3: "Web3") -> None:
        listening = await async_w3.async_net.listening

        assert is_boolean(listening)

    @pytest.mark.asyncio
    async def test_net_peer_count(self, async_w3: "Web3") -> None:
        peer_count = await async_w3.async_net.peer_count

        assert is_integer(peer_count)



import argparse
import asyncio
from collections import (
    defaultdict,
)
import logging
import sys
import timeit
from typing import (
    Any,
    Callable,
    Dict,
    Union,
)

from eth_typing import (
    ChecksumAddress,
)

from newchain_web3 import (
    AsyncHTTPProvider,
    HTTPProvider,
    Web3,
)
from newchain_web3.eth import (
    AsyncEth,
    Eth,
)
from newchain_web3.middleware import (
    async_buffered_gas_estimate_middleware,
    async_gas_price_strategy_middleware,
    buffered_gas_estimate_middleware,
    gas_price_strategy_middleware,
)
from newchain_web3.tools.benchmark.node import (
    GethBenchmarkFixture,
)
from newchain_web3.tools.benchmark.reporting import (
    print_entry,
    print_footer,
    print_header,
)
from newchain_web3.tools.benchmark.utils import (
    wait_for_aiohttp,
    wait_for_http,
)
from newchain_web3.types import (
    Wei,
)

KEYFILE_PW = "web3py-test"

parser = argparse.ArgumentParser()
parser.add_argument(
    "--num-calls",
    type=int,
    default=10,
    help="The number of RPC calls to make",
)

# TODO - layers to test:
# contract.functions.method(...).call()
# w3.eth.call(...)
# HTTPProvider.make_request(...)


def build_web3_http(endpoint_uri: str) -> Web3:
    wait_for_http(endpoint_uri)
    _w3 = Web3(
        HTTPProvider(endpoint_uri),
        middlewares=[gas_price_strategy_middleware, buffered_gas_estimate_middleware],
    )
    return _w3


async def build_async_w3_http(endpoint_uri: str) -> Web3:
    await wait_for_aiohttp(endpoint_uri)
    _w3 = Web3(
        AsyncHTTPProvider(endpoint_uri),
        middlewares=[
            async_gas_price_strategy_middleware,
            async_buffered_gas_estimate_middleware,
        ],
        modules={"eth": AsyncEth},
    )
    return _w3


def sync_benchmark(func: Callable[..., Any], n: int) -> Union[float, str]:
    try:
        starttime = timeit.default_timer()
        for _ in range(n):
            func()
        endtime = timeit.default_timer()
        execution_time = endtime - starttime
        return execution_time
    except Exception:
        return "N/A"


async def async_benchmark(func: Callable[..., Any], n: int) -> Union[float, str]:
    try:
        starttime = timeit.default_timer()
        for result in asyncio.as_completed([func() for _ in range(n)]):
            await result
        execution_time = timeit.default_timer() - starttime
        return execution_time
    except Exception:
        return "N/A"


def unlocked_account(w3: "Web3") -> ChecksumAddress:
    w3.geth.personal.unlock_account(w3.eth.coinbase, KEYFILE_PW)
    return w3.eth.coinbase


async def async_unlocked_account(w3: Web3, w3_eth: Eth) -> ChecksumAddress:
    # change w3_eth type to w3_eth: AsyncEth once AsyncEth reflects Eth
    coinbase = await w3_eth.coinbase  # type: ignore
    w3.geth.personal.unlock_account(coinbase, KEYFILE_PW)
    return coinbase


def main(logger: logging.Logger, num_calls: int) -> None:
    fixture = GethBenchmarkFixture()
    for built_fixture in fixture.build():
        for process in built_fixture:
            w3_http = build_web3_http(fixture.endpoint_uri)
            loop = asyncio.get_event_loop()
            async_w3_http = loop.run_until_complete(
                build_async_w3_http(fixture.endpoint_uri)
            )
            # TODO: swap out w3_http for the async_w3_http
            # once GethPersonal module is async
            async_unlocked_acct = loop.run_until_complete(
                async_unlocked_account(w3_http, async_w3_http.eth)
            )

            methods = [
                {
                    "name": "eth_gasPrice",
                    "params": {},
                    "exec": lambda: w3_http.eth.gas_price,
                    "async_exec": lambda: async_w3_http.eth.gas_price,
                },
                {
                    "name": "eth_sendTransaction",
                    "params": {},
                    "exec": lambda: w3_http.eth.send_transaction(
                        {
                            "to": "0xd3CdA913deB6f67967B99D67aCDFa1712C293601",
                            "from": unlocked_account(w3_http),
                            "value": Wei(12345),
                        }
                    ),
                    "async_exec": lambda: async_w3_http.eth.send_transaction(
                        {
                            "to": "0xd3CdA913deB6f67967B99D67aCDFa1712C293601",
                            "from": async_unlocked_acct,
                            "value": Wei(12345),
                        }
                    ),
                },
                {
                    "name": "eth_blockNumber",
                    "params": {},
                    "exec": lambda: w3_http.eth.block_number,
                    "async_exec": lambda: async_w3_http.eth.block_number,
                },
                {
                    "name": "eth_getBlock",
                    "params": {},
                    "exec": lambda: w3_http.eth.get_block(1),
                    "async_exec": lambda: async_w3_http.eth.get_block(1),
                },
            ]

            def benchmark(method: Dict[str, Any]) -> None:
                outcomes: Dict[str, Union[str, float]] = defaultdict(lambda: "N/A")
                outcomes["name"] = method["name"]
                outcomes["HTTPProvider"] = sync_benchmark(
                    method["exec"],
                    num_calls,
                )
                outcomes["AsyncHTTPProvider"] = loop.run_until_complete(
                    async_benchmark(method["async_exec"], num_calls)
                )
                print_entry(logger, outcomes)

            print_header(logger, num_calls)

            for method in methods:
                benchmark(method)

            print_footer(logger)


if __name__ == "__main__":
    args = parser.parse_args()

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.StreamHandler(sys.stdout))

    main(logger, args.num_calls)



from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Tuple,
)

from eth_utils import (
    is_same_address,
    to_bytes,
    to_tuple,
)
from eth_utils.toolz import (
    get_in,
)
from hexbytes import (
    HexBytes,
)

from ethpm.exceptions import (
    BytecodeLinkingError,
    EthPMValidationError,
)

if TYPE_CHECKING:
    from newchain_web3 import Web3  # noqa: F401


def get_linked_deployments(deployments: Dict[str, Any]) -> Dict[str, Any]:
    """
    Returns all deployments found in a chain URI's deployment data that
    contain link dependencies.
    """
    linked_deployments = {
        dep: data
        for dep, data in deployments.items()
        if get_in(("runtimeBytecode", "linkDependencies"), data)
    }
    for deployment, data in linked_deployments.items():
        if any(
            link_dep["value"] == deployment
            for link_dep in data["runtimeBytecode"]["linkDependencies"]
        ):
            raise BytecodeLinkingError(
                f"Link dependency found in {deployment} deployment that references its "
                "own contract instance, which is disallowed"
            )
    return linked_deployments


def validate_linked_references(
    link_deps: Tuple[Tuple[int, bytes], ...], bytecode: HexBytes
) -> None:
    """
    Validates that normalized linked_references (offset, expected_bytes)
    match the corresponding bytecode.
    """
    offsets, values = zip(*link_deps)
    for idx, offset in enumerate(offsets):
        value = values[idx]
        # https://github.com/python/mypy/issues/4975
        offset_value = int(offset)
        dep_length = len(value)
        end_of_bytes = offset_value + dep_length
        # Ignore b/c whitespace around ':' conflict b/w black & flake8
        actual_bytes = bytecode[offset_value:end_of_bytes]  # noqa: E203
        if actual_bytes != values[idx]:
            raise EthPMValidationError(
                "Error validating linked reference. "
                f"Offset: {offset} "
                f"Value: {values[idx]} "
                f"Bytecode: {bytecode!r} ."
            )


@to_tuple
def normalize_linked_references(
    data: List[Dict[str, Any]]
) -> Generator[Tuple[int, str, str], None, None]:
    """
    Return a tuple of information representing all insertions of a linked reference.
    (offset, type, value)
    """
    for deployment in data:
        for offset in deployment["offsets"]:
            yield offset, deployment["type"], deployment["value"]


def validate_deployments_tx_receipt(
    deployments: Dict[str, Any], w3: "Web3", allow_missing_data: bool = False
) -> None:
    """
    Validate that address and block hash found in deployment data match
    what is found on-chain. :allow_missing_data: by default, enforces
    validation of address and blockHash.
    """
    # todo: provide hook to lazily look up tx receipt via binary search if missing data
    for name, data in deployments.items():
        if "transaction" in data:
            tx_hash = data["transaction"]
            tx_receipt = w3.eth.get_transaction_receipt(tx_hash)
            # tx_address will be None if contract created via contract factory
            tx_address = tx_receipt["contractAddress"]

            if tx_address is None and allow_missing_data is False:
                raise EthPMValidationError(
                    "No contract address found in tx receipt. Unable to verify "
                    "address found in tx receipt matches address in manifest's "
                    "deployment data. If this validation is not necessary, "
                    "please enable `allow_missing_data` arg. "
                )

            if tx_address is not None and not is_same_address(
                tx_address, data["address"]
            ):
                raise EthPMValidationError(
                    f"Error validating tx_receipt for {name} deployment. "
                    f"Address found in manifest's deployment data: {data['address']} "
                    f"Does not match address found on tx_receipt: {tx_address}."
                )

            if "block" in data:
                if tx_receipt["blockHash"] != to_bytes(hexstr=data["block"]):
                    raise EthPMValidationError(
                        f"Error validating tx_receipt for {name} deployment. "
                        f"Block found in manifest's deployment data: {data['block']!r} "
                        "Does not match block found on "
                        f"tx_receipt: {tx_receipt['blockHash']!r}."
                    )
            elif allow_missing_data is False:
                raise EthPMValidationError(
                    "No block hash found in deployment data. "
                    "Unable to verify block hash on tx receipt. "
                    "If this validation is not necessary, please enable "
                    "`allow_missing_data` arg."
                )
        elif allow_missing_data is False:
            raise EthPMValidationError(
                "No transaction hash found in deployment data. "
                "Unable to validate tx_receipt. "
                "If this validation is not necessary, please "
                "enable `allow_missing_data` arg."
            )



from newchain_web3 import Web3
from newchain_web3.middleware import (
    geth_poa_middleware,
)
from newchain_web3.providers.auto import (
    load_provider_from_uri,
)

from .endpoints import (
    INFURA_GOERLI_DOMAIN,
    build_http_headers,
    build_infura_url,
)

_headers = build_http_headers()
_infura_url = build_infura_url(INFURA_GOERLI_DOMAIN)

w3 = Web3(load_provider_from_uri(_infura_url, _headers))
w3.middleware_onion.inject(geth_poa_middleware, layer=0)



from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)

from eth_utils import (
    combomethod,
    is_canonical_address,
    to_bytes,
    to_canonical_address,
)
from eth_utils.toolz import (
    assoc,
    curry,
    pipe,
)

from ethpm.exceptions import (
    BytecodeLinkingError,
    EthPMValidationError,
)
from ethpm.validation.misc import (
    validate_empty_bytes,
)
from newchain_web3._utils.validation import (
    validate_address,
)
from newchain_web3.contract import (
    Contract,
    ContractConstructor,
)

if TYPE_CHECKING:
    from newchain_web3 import Web3  # noqa: F401


class LinkableContract(Contract):
    """
    A subclass of web3.contract.Contract that is capable of handling
    contract factories with link references in their package's manifest.
    """

    unlinked_references: Optional[Tuple[Dict[str, Any]]] = None
    linked_references: Optional[Tuple[Dict[str, Any]]] = None
    needs_bytecode_linking = None

    def __init__(self, address: bytes, **kwargs: Any) -> None:
        if self.needs_bytecode_linking:
            raise BytecodeLinkingError(
                "Contract cannot be instantiated until its bytecode is linked."
            )
        validate_address(address)
        # type ignored to allow for undefined **kwargs on `Contract` base class __init__
        super().__init__(address=address, **kwargs)  # type: ignore

    @classmethod
    def factory(cls, w3: "Web3", class_name: str = None, **kwargs: Any) -> Contract:
        dep_link_refs = kwargs.get("unlinked_references")
        bytecode = kwargs.get("bytecode")
        needs_bytecode_linking = False
        if dep_link_refs and bytecode:
            if not is_prelinked_bytecode(to_bytes(hexstr=bytecode), dep_link_refs):
                needs_bytecode_linking = True
        kwargs = assoc(kwargs, "needs_bytecode_linking", needs_bytecode_linking)
        return super().factory(w3, class_name, **kwargs)

    @classmethod
    def constructor(cls, *args: Any, **kwargs: Any) -> ContractConstructor:
        if cls.needs_bytecode_linking:
            raise BytecodeLinkingError(
                "Contract cannot be deployed until its bytecode is linked."
            )
        return super().constructor(*args, **kwargs)

    @classmethod
    def link_bytecode(cls, attr_dict: Dict[str, str]) -> Type["LinkableContract"]:
        """
        Return a cloned contract factory with the deployment / runtime bytecode linked.

        :attr_dict: Dict[`ContractType`: `Address`] for all deployment and runtime
        link references.
        """
        if not cls.unlinked_references and not cls.linked_references:
            raise BytecodeLinkingError("Contract factory has no linkable bytecode.")
        if not cls.needs_bytecode_linking:
            raise BytecodeLinkingError(
                "Bytecode for this contract factory does not require bytecode linking."
            )
        cls.validate_attr_dict(attr_dict)
        bytecode = apply_all_link_refs(cls.bytecode, cls.unlinked_references, attr_dict)
        runtime = apply_all_link_refs(
            cls.bytecode_runtime, cls.linked_references, attr_dict
        )
        linked_class = cls.factory(cls.w3, bytecode_runtime=runtime, bytecode=bytecode)
        if linked_class.needs_bytecode_linking:
            raise BytecodeLinkingError(
                "Expected class to be fully linked, but class "
                "still needs bytecode linking."
            )
        return linked_class

    @combomethod
    def validate_attr_dict(self, attr_dict: Dict[str, str]) -> None:
        """
        Validates that ContractType keys in attr_dict reference existing
        manifest ContractTypes.
        """
        attr_dict_names = attr_dict.keys()

        if not self.unlinked_references and not self.linked_references:
            raise BytecodeLinkingError(
                "Unable to validate attr dict, this contract has no "
                "linked/unlinked references."
            )

        unlinked_refs = self.unlinked_references or ({},)
        linked_refs = self.linked_references or ({},)
        all_link_refs = unlinked_refs + linked_refs

        all_link_names = {ref["name"] for ref in all_link_refs if ref}
        if attr_dict_names != all_link_names:
            raise BytecodeLinkingError(
                "All link references must be defined when calling "
                "`link_bytecode` on a contract factory."
            )
        for address in attr_dict.values():
            validate_address(address)


def is_prelinked_bytecode(bytecode: bytes, link_refs: List[Dict[str, Any]]) -> bool:
    """
    Returns False if all expected link_refs are unlinked, otherwise returns True.
    todo support partially pre-linked bytecode (currently all or nothing)
    """
    for link_ref in link_refs:
        for offset in link_ref["offsets"]:
            try:
                validate_empty_bytes(offset, link_ref["length"], bytecode)
            except EthPMValidationError:
                return True
    return False


def apply_all_link_refs(
    bytecode: bytes, link_refs: List[Dict[str, Any]], attr_dict: Dict[str, str]
) -> bytes:
    """
    Applies all link references corresponding to a valid attr_dict to the bytecode.
    """
    if link_refs is None:
        return bytecode
    link_fns = (
        apply_link_ref(offset, ref["length"], attr_dict[ref["name"]])
        for ref in link_refs
        for offset in ref["offsets"]
    )
    linked_bytecode = pipe(bytecode, *link_fns)
    return linked_bytecode


@curry
def apply_link_ref(offset: int, length: int, value: bytes, bytecode: bytes) -> bytes:
    """
    Returns the new bytecode with `value` put into the location indicated by
    `offset` and `length`.
    """
    try:
        validate_empty_bytes(offset, length, bytecode)
    except EthPMValidationError:
        raise BytecodeLinkingError("Link references cannot be applied to bytecode")

    address = value if is_canonical_address(value) else to_canonical_address(value)
    new_bytes = (
        # Ignore linting error b/c conflict b/w black & flake8
        bytecode[:offset]
        + address
        + bytecode[offset + length :]  # noqa: E201, E203
    )
    return new_bytes



import json
import pytest
from typing import (
    TYPE_CHECKING,
    cast,
)

from eth_typing import (
    ChecksumAddress,
)
from eth_utils import (
    is_checksum_address,
    is_list_like,
    is_same_address,
    is_string,
)
from hexbytes import (
    HexBytes,
)

from newchain_web3 import (
    constants,
)
from newchain_web3.datastructures import (
    AttributeDict,
)
from newchain_web3.types import (  # noqa: F401
    TxParams,
    Wei,
)

if TYPE_CHECKING:
    from newchain_web3 import Web3  # noqa: F401

PRIVATE_KEY_HEX = "0x56ebb41875ceedd42e395f730e03b5c44989393c9f0484ee6bc05f933673458f"
SECOND_PRIVATE_KEY_HEX = (
    "0x56ebb41875ceedd42e395f730e03b5c44989393c9f0484ee6bc05f9336712345"
)
THIRD_PRIVATE_KEY_HEX = (
    "0x56ebb41875ceedd42e395f730e03b5c44989393c9f0484ee6bc05f9336754321"
)
PASSWORD = "web3-testing"
ADDRESS = "0x844B417c0C58B02c2224306047B9fb0D3264fE8c"
SECOND_ADDRESS = "0xB96b6B21053e67BA59907E252D990C71742c41B8"


PRIVATE_KEY_FOR_UNLOCK = (
    "0x392f63a79b1ff8774845f3fa69de4a13800a59e7083f5187f1558f0797ad0f01"
)
ACCOUNT_FOR_UNLOCK = "0x12efDc31B1a8FA1A1e756DFD8A1601055C971E13"


class GoEthereumPersonalModuleTest:
    def test_personal_import_raw_key(self, w3: "Web3") -> None:
        actual = w3.geth.personal.import_raw_key(PRIVATE_KEY_HEX, PASSWORD)
        assert actual == ADDRESS

    def test_personal_list_accounts(self, w3: "Web3") -> None:
        accounts = w3.geth.personal.list_accounts()
        assert is_list_like(accounts)
        assert len(accounts) > 0
        assert all((is_checksum_address(item) for item in accounts))

    def test_personal_list_wallets(self, w3: "Web3") -> None:
        wallets = w3.geth.personal.list_wallets()
        assert is_list_like(wallets)
        assert len(wallets) > 0
        assert is_checksum_address(wallets[0]["accounts"][0]["address"])
        assert is_string(wallets[0]["accounts"][0]["url"])
        assert is_string(wallets[0]["status"])
        assert is_string(wallets[0]["url"])

    def test_personal_lock_account(
        self, w3: "Web3", unlockable_account_dual_type: ChecksumAddress
    ) -> None:
        # TODO: how do we test this better?
        w3.geth.personal.lock_account(unlockable_account_dual_type)

    def test_personal_unlock_account_success(
        self,
        w3: "Web3",
        unlockable_account_dual_type: ChecksumAddress,
        unlockable_account_pw: str,
    ) -> None:
        result = w3.geth.personal.unlock_account(
            unlockable_account_dual_type, unlockable_account_pw
        )
        assert result is True

    def test_personal_unlock_account_failure(
        self, w3: "Web3", unlockable_account_dual_type: ChecksumAddress
    ) -> None:
        with pytest.raises(ValueError):
            w3.geth.personal.unlock_account(
                unlockable_account_dual_type, "bad-password"
            )

    def test_personal_new_account(self, w3: "Web3") -> None:
        new_account = w3.geth.personal.new_account(PASSWORD)
        assert is_checksum_address(new_account)

    def test_personal_send_transaction(
        self,
        w3: "Web3",
        unlockable_account_dual_type: ChecksumAddress,
        unlockable_account_pw: str,
    ) -> None:
        assert (
            w3.eth.get_balance(unlockable_account_dual_type) > constants.WEI_PER_ETHER
        )
        txn_params: TxParams = {
            "from": unlockable_account_dual_type,
            "to": unlockable_account_dual_type,
            "gas": 21000,
            "value": Wei(1),
            "gasPrice": w3.toWei(1, "gwei"),
        }
        txn_hash = w3.geth.personal.send_transaction(txn_params, unlockable_account_pw)
        assert txn_hash
        transaction = w3.eth.get_transaction(txn_hash)

        assert is_same_address(
            transaction["from"], cast(ChecksumAddress, txn_params["from"])
        )
        assert is_same_address(
            transaction["to"], cast(ChecksumAddress, txn_params["to"])
        )
        assert transaction["gas"] == txn_params["gas"]
        assert transaction["value"] == txn_params["value"]
        assert transaction["gasPrice"] == txn_params["gasPrice"]

    def test_personal_sign_and_ecrecover(
        self,
        w3: "Web3",
        unlockable_account_dual_type: ChecksumAddress,
        unlockable_account_pw: str,
    ) -> None:
        message = "test-web3-geth-personal-sign"
        signature = w3.geth.personal.sign(
            message, unlockable_account_dual_type, unlockable_account_pw
        )
        signer = w3.geth.personal.ec_recover(message, signature)
        assert is_same_address(signer, unlockable_account_dual_type)

    @pytest.mark.xfail(
        reason="personal_sign_typed_data JSON RPC call has not been released in geth"
    )
    def test_personal_sign_typed_data(
        self,
        w3: "Web3",
        unlockable_account_dual_type: ChecksumAddress,
        unlockable_account_pw: str,
    ) -> None:
        typed_message = """
            {
                "types": {
                    "EIP712Domain": [
                        {"name": "name", "type": "string"},
                        {"name": "version", "type": "string"},
                        {"name": "chainId", "type": "uint256"},
                        {"name": "verifyingContract", "type": "address"}
                    ],
                    "Person": [
                        {"name": "name", "type": "string"},
                        {"name": "wallet", "type": "address"}
                    ],
                    "Mail": [
                        {"name": "from", "type": "Person"},
                        {"name": "to", "type": "Person"},
                        {"name": "contents", "type": "string"}
                    ]
                },
                "primaryType": "Mail",
                "domain": {
                    "name": "Ether Mail",
                    "version": "1",
                    "chainId": "0x01",
                    "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC"
                },
                "message": {
                    "from": {
                        "name": "Cow",
                        "wallet": "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826"
                    },
                    "to": {
                        "name": "Bob",
                        "wallet": "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB"
                    },
                    "contents": "Hello, Bob!"
                }
            }
        """
        signature = HexBytes(
            w3.geth.personal.sign_typed_data(
                json.loads(typed_message),
                unlockable_account_dual_type,
                unlockable_account_pw,
            )
        )

        expected_signature = HexBytes(
            "0xc8b56aaeefd10ab4005c2455daf28d9082af661ac347cd"
            "b612d5b5e11f339f2055be831bf57a6e6cb5f6d93448fa35"
            "c1bd56fe1d745ffa101e74697108668c401c"
        )
        assert signature == expected_signature
        assert len(signature) == 32 + 32 + 1


class GoEthereumAsyncPersonalModuleTest:
    @pytest.mark.asyncio
    async def test_async_sign_and_ec_recover(
        self,
        async_w3: "Web3",
        unlockable_account_dual_type: ChecksumAddress,
        unlockable_account_pw: str,
    ) -> None:
        message = "This is a test"
        signature = await async_w3.geth.personal.sign(
            message, unlockable_account_dual_type, unlockable_account_pw  # type: ignore
        )
        address = await async_w3.geth.personal.ec_recover(  # type: ignore
            message, signature
        )
        assert is_same_address(unlockable_account_dual_type, address)

    @pytest.mark.asyncio
    async def test_async_import_key(self, async_w3: "Web3") -> None:
        address = await async_w3.geth.personal.import_raw_key(
            THIRD_PRIVATE_KEY_HEX, "Testing"  # type: ignore
        )
        assert address is not None

    @pytest.mark.asyncio
    async def test_async_list_accounts(self, async_w3: "Web3") -> None:
        accounts = await async_w3.geth.personal.list_accounts()  # type: ignore
        assert len(accounts) > 0

    @pytest.mark.asyncio
    async def test_async_list_wallets(self, async_w3: "Web3") -> None:
        wallets = await async_w3.geth.personal.list_wallets()  # type: ignore
        assert isinstance(wallets[0], AttributeDict)

    @pytest.mark.asyncio
    async def test_async_new_account(self, async_w3: "Web3") -> None:
        passphrase = "Create New Account"
        account = await async_w3.geth.personal.new_account(passphrase)  # type: ignore
        assert is_checksum_address(account)

    @pytest.mark.asyncio
    async def test_async_unlock_lock_account(
        self,
        async_w3: "Web3",
        unlockable_account_dual_type: ChecksumAddress,
        unlockable_account_pw: str,
    ) -> None:
        unlocked = await async_w3.geth.personal.unlock_account(  # type: ignore
            unlockable_account_dual_type, unlockable_account_pw
        )
        assert unlocked is True
        locked = await async_w3.geth.personal.lock_account(  # type: ignore
            unlockable_account_dual_type
        )
        assert locked is True

    @pytest.mark.asyncio
    async def test_async_send_transaction(
        self,
        async_w3: "Web3",
        unlockable_account_dual_type: ChecksumAddress,
        unlockable_account_pw: str,
    ) -> None:
        tx_params = TxParams()
        tx_params["to"] = unlockable_account_dual_type
        tx_params["from"] = unlockable_account_dual_type
        tx_params["value"] = Wei(123)
        response = await async_w3.geth.personal.send_transaction(  # type: ignore
            tx_params, unlockable_account_pw
        )
        assert response is not None

    @pytest.mark.xfail(
        reason="personal_signTypedData JSON RPC call has not been released in geth"
    )
    @pytest.mark.asyncio
    async def test_async_sign_typed_data(
        self,
        async_w3: "Web3",
        unlockable_account_dual_type: ChecksumAddress,
        unlockable_account_pw: str,
    ) -> None:
        message = {"message": "This is a test"}
        signature = await async_w3.geth.personal.sign_typed_data(
            message, unlockable_account_dual_type, unlockable_account_pw  # type: ignore
        )
        address = await async_w3.geth.personal.ec_recover(
            message, signature  # type: ignore
        )
        assert is_same_address(unlockable_account_dual_type, address)



import functools
import threading
from typing import (
    Any,
    Callable,
    TypeVar,
    cast,
)
import warnings

TFunc = TypeVar("TFunc", bound=Callable[..., Any])


def reject_recursive_repeats(to_wrap: Callable[..., Any]) -> Callable[..., Any]:
    """
    Prevent simple cycles by returning None when called recursively with same instance
    """
    # types ignored b/c dynamically set attribute
    to_wrap.__already_called = {}  # type: ignore

    @functools.wraps(to_wrap)
    def wrapped(*args: Any) -> Any:
        arg_instances = tuple(map(id, args))
        thread_id = threading.get_ident()
        thread_local_args = (thread_id,) + arg_instances
        if thread_local_args in to_wrap.__already_called:  # type: ignore
            raise ValueError(f"Recursively called {to_wrap} with {args!r}")
        to_wrap.__already_called[thread_local_args] = True  # type: ignore
        try:
            wrapped_val = to_wrap(*args)
        finally:
            del to_wrap.__already_called[thread_local_args]  # type: ignore
        return wrapped_val

    return wrapped


def deprecated_for(replace_message: str) -> Callable[..., Any]:
    """
    Decorate a deprecated function, with info about what to use instead, like:

    @deprecated_for("toBytes()")
    def toAscii(arg):
        ...
    """

    def decorator(to_wrap: TFunc) -> TFunc:
        @functools.wraps(to_wrap)
        def wrapper(*args: Any, **kwargs: Any) -> Callable[..., Any]:
            warnings.warn(
                f"{to_wrap.__name__} is deprecated in favor of {replace_message}",
                category=DeprecationWarning,
            )
            return to_wrap(*args, **kwargs)

        return cast(TFunc, wrapper)

    return decorator



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
)

from eth_utils.curried import (
    apply_formatter_if,
    apply_formatters_to_dict,
    apply_key_map,
    is_null,
)
from eth_utils.toolz import (
    complement,
    compose,
)
from hexbytes import (
    HexBytes,
)

# from newchain_web3._utils.toolz import (
#     complement,
#     compose,
# )
from newchain_web3._utils.rpc_abi import (
    RPC,
)
from newchain_web3.middleware.formatting import (
    async_construct_formatting_middleware,
    construct_formatting_middleware,
)
from newchain_web3.types import (
    AsyncMiddleware,
    RPCEndpoint,
)

if TYPE_CHECKING:
    from newchain_web3 import Web3  # noqa: F401

is_not_null = complement(is_null)

remap_geth_poa_fields = apply_key_map(
    {
        "extraData": "proofOfAuthorityData",
    }
)

pythonic_geth_poa = apply_formatters_to_dict(
    {
        "proofOfAuthorityData": HexBytes,
    }
)

geth_poa_cleanup = compose(pythonic_geth_poa, remap_geth_poa_fields)


geth_poa_middleware = construct_formatting_middleware(
    result_formatters={
        RPC.eth_getBlockByHash: apply_formatter_if(is_not_null, geth_poa_cleanup),
        RPC.eth_getBlockByNumber: apply_formatter_if(is_not_null, geth_poa_cleanup),
    },
)


async def async_geth_poa_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], w3: "Web3"
) -> AsyncMiddleware:
    middleware = await async_construct_formatting_middleware(
        result_formatters={
            RPC.eth_getBlockByHash: apply_formatter_if(is_not_null, geth_poa_cleanup),
            RPC.eth_getBlockByNumber: apply_formatter_if(is_not_null, geth_poa_cleanup),
        },
    )
    return await middleware(make_request, w3)



import idna


class AddressMismatch(ValueError):
    """
    In order to set up reverse resolution correctly, the ENS name should first
    point to the address. This exception is raised if the name does
    not currently point to the address.
    """

    pass


class InvalidName(idna.IDNAError):
    """
    This exception is raised if the provided name does not meet
    the syntax standards specified in `EIP 137 name syntax
    <https://github.com/ethereum/EIPs/blob/master/EIPS/eip-137.md#name-syntax>`_.

    For example: names may not start with a dot, or include a space.
    """

    pass


class UnauthorizedError(Exception):
    """
    Raised if the sending account is not the owner of the name
    you are trying to modify. Make sure to set ``from`` in the
    ``transact`` keyword argument to the owner of the name.
    """

    pass


class UnownedName(Exception):
    """
    Raised if you are trying to modify a name that no one owns.

    If working on a subdomain, make sure the subdomain gets created
    first with :meth:`~ens.ENS.setup_address`.
    """

    pass


class ResolverNotFound(Exception):
    """
    Raised if no resolver was found for the name you are trying to resolve.
    """

    pass


class UnsupportedFunction(Exception):
    """
    Raised if a resolver does not support a particular method.
    """

    pass


class BidTooLow(ValueError):
    """
    Raised if you bid less than the minimum amount
    """

    pass


class InvalidBidHash(ValueError):
    """
    Raised if you supply incorrect data to generate the bid hash.
    """

    pass


class InvalidLabel(ValueError):
    """
    Raised if you supply an invalid label
    """

    pass


class OversizeTransaction(ValueError):
    """
    Raised if a transaction you are trying to create would cost so
    much gas that it could not fit in a block.

    For example: when you try to start too many auctions at once.
    """

    pass


class UnderfundedBid(ValueError):
    """
    Raised if you send less wei with your bid than you declared
    as your intent to bid.
    """

    pass



from typing import (
    Any,
    Dict,
    ItemsView,
    List,
)

from eth_typing import (
    Address,
    HexStr,
)

from ethpm.validation.package import (
    validate_contract_name,
)
from newchain_web3._utils.compat import (
    TypedDict,
)
from newchain_web3.eth import (
    Contract,
)


class Deployments:
    """
    Deployment object to access instances of
    deployed contracts belonging to a package.
    """

    def __init__(
        self,
        deployment_data: Dict[str, Dict[str, str]],
        contract_instances: Dict[str, Contract],
    ) -> None:
        self.deployment_data = deployment_data
        self.contract_instances = contract_instances

    def __getitem__(self, key: str) -> Dict[str, str]:
        return self.get(key)

    def __contains__(self, key: str) -> bool:
        return key in self.deployment_data

    def get(self, key: str) -> Dict[str, str]:
        self._validate_name_and_references(key)
        return self.deployment_data[key]

    def items(self) -> ItemsView[str, Dict[str, str]]:
        item_dict = {name: self.get(name) for name in self.deployment_data}
        return item_dict.items()

    def values(self) -> List[Dict[str, str]]:
        values = [self.get(name) for name in self.deployment_data]
        return values

    def get_instance(self, contract_name: str) -> Contract:
        """
        Fetches a contract instance belonging to deployment
        after validating contract name.
        """
        self._validate_name_and_references(contract_name)
        return self.contract_instances[contract_name]

    def _validate_name_and_references(self, name: str) -> None:
        validate_contract_name(name)
        if name not in self.deployment_data:
            raise KeyError(
                f"Contract deployment: {name} not found in deployment data. "
                "Available deployments include: "
                f"{list(sorted(self.deployment_data.keys()))}."
            )


class DeploymentData(TypedDict):
    address: Address
    transaction: HexStr
    block: HexStr
    runtime_bytecode: Dict[str, Any]
    compiler: Dict[str, str]
    contractType: str
